﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace XamSyntaxEditorInvalidVars
{
    public static class Extensions
    {

        public static List<string> GetVariables(this string templateString)
        {
            if (String.IsNullOrWhiteSpace(templateString))
            {
                return new List<string>();
            }
            List<string> toReturn = new List<string>();
            bool encounteredDelimiter = false;
            bool isInExpression = false;
            StringBuilder currentExpression = new StringBuilder();
            foreach (char currentchar in templateString)
            {
                if (currentchar == '%')
                {
                    if (isInExpression)
                    {
                        isInExpression = false;
                        string expression = currentExpression.ToString();
                        toReturn = toReturn.Concat(GetVariablesFromExpression(expression)).ToList();
                        currentExpression.Clear();
                    }
                    else {
                        encounteredDelimiter = !encounteredDelimiter;
                    }
                }
                else { //we encoutered something else
                    if (encounteredDelimiter)
                    {
                        isInExpression = true;
                        encounteredDelimiter = false;
                        currentExpression.Clear();
                        currentExpression.Append(currentchar);
                    }
                    else {
                        if (isInExpression)
                        {
                            currentExpression.Append(currentchar);
                        }
                    }
                }
            }

            return toReturn;
        }

        private static readonly Regex variableRegex = new Regex("^[0-9A-Za-z_. -\\[\\]]*$");
        private static List<string> GetVariablesFromExpression(string expression)
        {
            List<string> toReturn = new List<string>();
            string trimmedExpression = expression.Trim();
            if (variableRegex.IsMatch(trimmedExpression))
            {
                toReturn.Add(trimmedExpression);
            }
            else {
                //We're going to parse this as a function
                //Do a simple PDA (push down automata) to properly parse stuff 
                Stack<object> pdaStack = new Stack<object>();
                List<string> arguments = new List<string>();
                StringBuilder currentFunctionName = new StringBuilder();
                StringBuilder currentArgument = new StringBuilder();
                bool stop = false;
                foreach (char c in trimmedExpression)
                {
                    //This means we've "fallen off the end" of the string... throw an exception
                    if (stop)
                    {
                        throw new Exception(String.Format("Expected '%' but encountered extra text after function '{0}'.", currentFunctionName));
                    }
                    switch (c)
                    {
                        case '(':
                            //Count of >0 means that this is all part of the argument to the function
                            if (pdaStack.Count != 0)
                            {
                                currentArgument.Append(c);
                            }
                            pdaStack.Push(new object());
                            break;
                        case ')':
                            pdaStack.Pop();
                            if (pdaStack.Count != 0)
                            {
                                currentArgument.Append(c);
                            }
                            else {
                                stop = true;
                                arguments.Add(currentArgument.ToString());
                            }
                            break;
                        case ',':
                            //If the stack is only one element high, we are in the top-level function... store the argument and reset, then continue
                            if (pdaStack.Count == 1)
                            {
                                arguments.Add(currentArgument.ToString());
                                currentArgument.Clear();
                            }
                            else {
                                //Keep going
                                currentArgument.Append(c);
                            }
                            break;
                        default:
                            if (pdaStack.Count == 0)
                            {
                                currentFunctionName.Append(c);
                            }
                            else {
                                currentArgument.Append(c);
                            }
                            break;
                    }
                }

                if (arguments.Count == 0 && currentFunctionName.Length > 0)
                {
                    toReturn.Add(currentFunctionName.ToString());
                }
                else
                {
                    toReturn = arguments.Aggregate(toReturn, (current, argument) => current.Concat(GetVariablesFromExpression(argument)).ToList());
                    if (pdaStack.Count > 0)
                    {//Make sure we aren't expecting more parens
                        throw new Exception("Expected ')', but end of expression reached.  Please ensure all your function declarations are closed.");
                    }
                }
            }
            return toReturn;
        }

        public static List<string> ReturnInvalidVars(this string text, List<string> ValidVars)
        {
            var toReturn = new List<string>();
            text.GetVariables().ToList().ForEach(var =>
            {
                if (!ValidVars.Contains(var, StringComparer.InvariantCultureIgnoreCase))
                {
                    toReturn.Add(var);
                }

            });
            return toReturn;
        }

        public static string Flatten(this IEnumerable elems, string separator)
        {
            if (elems == null)
            {
                return null;
            }

            StringBuilder sb = new StringBuilder();
            foreach (object elem in elems)
            {
                if (sb.Length > 0)
                {
                    sb.Append(separator);
                }

                sb.Append(elem);
            }

            return sb.ToString();
        }
    }
}
