﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Media;
using Infragistics.Controls.Editors;
using Infragistics.Documents;
using Infragistics.Documents.Parsing;

namespace XamSyntaxEditorInvalidVars
{
    public class MainWindowViewModel : ViewModelBase
    {
        public MainWindowViewModel()
        {

        }



        private string _selectedListBoxItem;
        public string SelectedListBoxItem
        {
            get
            {
                return _selectedListBoxItem;
            }
            set
            {
                _selectedListBoxItem = value;
                RaisePropertyChanged(() => SelectedListBoxItem);
            }
        }

        private List<string> _jobVariables;
        public List<string> JobVariables
        {
            get
            {
                if (_jobVariables == null)
                {
                    _jobVariables = new List<string>();
                    for(int j = 0; j <= 500; j++)
                    {
                        _jobVariables.Add("LASTRUN_" + j);
                    }


                }
                return _jobVariables;
            }
        }

        private ClassificationAppearanceMap _classificationAppearanceMap;

        public ClassificationAppearanceMap ClassificationAppearanceMap
        {
            get { return _classificationAppearanceMap; }
            set
            {
                _classificationAppearanceMap = value;
                RaisePropertyChanged(() => ClassificationAppearanceMap);
            }
        }

        private TextDocument _textDocument;

        public TextDocument TextDocument
        {
            get
            {
                if (_textDocument == null)
                {
                    _textDocument = new TextDocument();
                    _textDocument.AutoCasingMode = TextDocumentAutoCasingMode.None;
                    _textDocument.TextChanged += TdOnTextChanged;
                    _textDocument.InitializeText(QueryText);
                    CreateCustomGrammar(ReturnInvalidVars());

                }

                return _textDocument;

            }
            set
            {
                _textDocument = value;
                RaisePropertyChanged(() => TextDocument);
            }

        }

        private void TdOnTextChanged(object sender, TextChangedEventArgs e)
         {

            //QueryText = ((TextDocument)sender).CurrentSnapshot.GetText();
            //var tl = edv.SyntaxEditor.Caret.TextLocation;
            //var tl2 = new TextLocation(tl.Line, tl.Character + 1);
            

            //TextLocation tl = this.xamSyntaxEditor1.Caret.TextLocation;
            //caret.TextLocationNormalized.Line = 2;
            //edv.SyntaxEditor.Caret.TextLocationNormalized.Line = 2;
            
            ////CreateCustomGrammar(ReturnInvalidVars());

            //SnapshotPoint ssp = SnapshotPoint.FromTextLocation(tl2, TextDocument.CurrentSnapshot);
            //edv.SelectionManager.SelectionStart = ssp;
            //edv.SelectionManager.SelectionEnd = ssp;

        }

        private List<string> ReturnInvalidVars()
        {
            var toReturn = new List<string>();
            QueryText.GetVariables().ToList().ForEach(var =>
            {
                if (!JobVariables.Contains(var, StringComparer.InvariantCultureIgnoreCase))
                {
                    toReturn.Add(var);
                }
                
            });
            return toReturn;
        }

        public EditorDocumentView edv;

        private string _queryText;
        public string QueryText
        {
            get
            {
                if (_queryText == null)
                {
                    _queryText = "\n\n\n\n\n\n\n\n\n Select * from Orders \n\n\n\n\n\n\n\n\n";
                }

                return _queryText;
            }
            set { _queryText = value; }
        }

        public string CreateValidVarRegEx()
        {
            var returnString = string.Empty;
            var startReg = "(?i)(";
            var varReg = string.Empty;
            var pipe = "|";
            var endReg = ")(?-i)";

            varReg = JobVariables.Flatten("|");

            returnString = string.Format("{0}{1}{2}", startReg, varReg, endReg );


            return returnString;
        }

        private void CreateCustomGrammar(List<string> invalidVars)
        {

            try
            {
                ClassificationAppearanceMap = null;

                var tsqlLanguage = TSqlLanguage.Instance;
                var customGrammar2 = tsqlLanguage.Grammar.Clone();
                ModLexerStates2(customGrammar2);

                var customLanguage = new CustomLanguage(customGrammar2);

                foreach (var serviceKey in tsqlLanguage.ServicesManager.GetRegisteredServiceKeys())
                    customLanguage.ServicesManager.RegisterService(serviceKey, tsqlLanguage.ServicesManager.GetService(serviceKey));

                foreach (var serviceType in tsqlLanguage.ServicesManager.GetRegisteredServiceTypes())
                    customLanguage.ServicesManager.RegisterService(serviceType, tsqlLanguage.ServicesManager.GetService(serviceType));

                TextDocument.Language = customLanguage;
            }
            catch (Exception e)
            {
                
            }
        }

        public Grammar ModLexerStates2(Grammar grammar)
        {
            #region grammar region

            NonTerminalSymbol identifierResolved;
            if (grammar.NonTerminalSymbols.TryGetSymbol("_identifierResolved", out identifierResolved) == false)
            {
                Debug.Fail("Cannot find the _identifierResolved symbol.");
                return null;
            }

            var alternateRule = identifierResolved.Rule as AlternationSyntaxRule;
            if (alternateRule == null)
            {
                Debug.Fail("The _identifierResolved symbol does not have an alternation at the root level rule, as expected.");
                return null;
            }

            #endregion

            var defaultLexerState = grammar.LexerStates.DefaultLexerState;
            var languageName = "Custom";
            int j = 0;

            // Not sure of exactly how many of these cause an out of memory exception, but 750 will definitely cause the
            // out of memory exception

            JobVariables.ForEach(var =>
            {
                var validVar = defaultLexerState.Symbols.Insert(0, "ValidVar" + j++.ToString(), "%" + var + "%", TerminalSymbolComparison.LiteralIgnoreCase);
                validVar.LanguageElementName = languageName;
                alternateRule.Rules.Add(validVar);
            });

            if (ClassificationAppearanceMap == null)
                ClassificationAppearanceMap = new ClassificationAppearanceMap();

            ClassificationAppearanceMap.AddMapEntry(
                new ClassificationType(languageName, languageName, "The keyword for the Custom name"),
                new TextDocumentAppearance() { Foreground = new SolidColorBrush(Colors.DarkOrange), FontBold = true });

            return grammar;
            
        }

        public void DAVariableLanguage2(Grammar grammar)
        {

            var languageElementName = "Custom";
            //define DAVarStart TerminalSymbol
            var daVarStart = grammar.LexerStates.DefaultLexerState.Symbols.Add("DaVarStart", "%");
            daVarStart.LanguageElementName = languageElementName;

            //define DaVarContent TerminalSymbol
            var lexerState = grammar.LexerStates.Add("DAVAR");
            daVarStart.LexerStateToEnter = lexerState;

            var daVarContent = lexerState.Symbols.Add("DaVarContent", @"(LASTRUN)", TerminalSymbolComparison.RegularExpression);
            daVarContent.LanguageElementName = languageElementName;
            //daVarContent.LookaheadPattern = @"[^\S]";
            //daVarContent.IsLookaheadNegative = true;

            //var daVarContent2 = lexerState.Symbols.Add("DaVarContent2", "%+", TerminalSymbolComparison.RegularExpression);
            //daVarContent2.LookaheadPattern = "/";
            //daVarContent2.LanguageElementName = languageElementName;

            //define DaVarContent TerminalSymbol
            var daVarEnd = lexerState.Symbols.Add("DaVarEnd", "%", TerminalSymbolComparison.Literal, true);
            daVarEnd.LanguageElementName = languageElementName;

            var variableDeclarationStatement =
                    grammar.NonTerminalSymbols.Add("VariableDeclarationStatement");

            variableDeclarationStatement.Rule = new EmptySyntaxRule();

            grammar.StartSymbol = variableDeclarationStatement;

            if (ClassificationAppearanceMap == null)
                ClassificationAppearanceMap = new ClassificationAppearanceMap();

            ClassificationAppearanceMap.AddMapEntry(
                new ClassificationType(languageElementName, daVarStart + " Name", "The keyword for the " + daVarStart + " name"),
                new TextDocumentAppearance() { Foreground = new SolidColorBrush(Colors.Chartreuse), FontBold = true });

        }

        public Grammar AlternationExample()
        {
            var grammar = new Grammar();
            var defaultLexerState = grammar.LexerStates.DefaultLexerState;
            // Include the built-in symbols which should be matched in the default lexer state.
            defaultLexerState.Symbols.Add(grammar.WhitespaceSymbol);
            defaultLexerState.Symbols.Add(grammar.NewLineSymbol);
            var equals = defaultLexerState.Symbols.Add("EqualsToken", "=");
            var semicolon = defaultLexerState.Symbols.Add("SemicolonToken", ";");
            var varKeyword = defaultLexerState.Symbols.Add("VarKeyword", "var");
            var identifier = defaultLexerState.Symbols.Add("IdentifierToken",
                "[_a-zA-Z][_a-zA-Z0-9]*", TerminalSymbolComparison.RegularExpression);
            var type = grammar.NonTerminalSymbols.Add("Type");
            type.Rule = identifier; // Implicit conversion
            var expression = grammar.NonTerminalSymbols.Add("Expression");
            expression.Rule = new EmptySyntaxRule();
            var variableDeclarationStatement =
                grammar.NonTerminalSymbols.Add("VariableDeclarationStatement");
            variableDeclarationStatement.Rule =
                new AlternationSyntaxRule(
                    new ConcatenationSyntaxRule(
                        type, // Implicit conversions
                        identifier,
                        new OptionalSyntaxRule(
                            new ConcatenationSyntaxRule(
                                equals,
                                expression
                            )
                        ),
                        semicolon
                    ),
                    new ConcatenationSyntaxRule(
                        varKeyword,
                        identifier,
                        equals,
                        expression,
                        semicolon
                    )
                );
            grammar.StartSymbol = variableDeclarationStatement;
            return grammar;
        }

        //Satu work below
        internal void DAVariableLanguage(Grammar grammar)
        {
            //define DAVarStart TerminalSymbol
            var daVarStart = grammar.LexerStates.DefaultLexerState.Symbols.Add("DaVarStart", "%");
            daVarStart.LanguageElementName = "Custom";
            //define DaVarContent TerminalSymbol
            daVarStart.LexerStateToEnter = grammar.LexerStates.Add("DAVAR");
            var daVarContent = daVarStart.LexerStateToEnter.Symbols.Add("DaVarContent", "[^%]+", TerminalSymbolComparison.RegularExpression);
            daVarContent.LookaheadPattern = "/";
            daVarContent.LanguageElementName = "Custom";
            //define DaVarContent TerminalSymbol
            var daVarEnd = daVarStart.LexerStateToEnter.Symbols.Add("DaVarEnd", "%", TerminalSymbolComparison.Literal, true);
            daVarEnd.LanguageElementName = "Custom";



        }

        public void AddIdentifierSymbol(Grammar grammar, string identifier, string invalidVar, Color foreground)
        {
            var gram = grammar;
            
            #region hide region

            NonTerminalSymbol identifierResolved;
            if (grammar.NonTerminalSymbols.TryGetSymbol("_identifierResolved", out identifierResolved) == false)
            {
                Debug.Fail("Cannot find the _identifierResolved symbol.");
                return;
            }

            var alternateRule = identifierResolved.Rule as AlternationSyntaxRule;
            if (alternateRule == null)
            {
                Debug.Fail("The _identifierResolved symbol does not have an alternation at the root level rule, as expected.");
                return;
            }

            #endregion

            // Create a terminal symbol to recognize the identifier and insert it at the beginning of the default lexer state so
            // it gets priority over the IdentifierToken symbol, which will also match the identifier.
            //var identifierSymbol = grammar.LexerStates.DefaultLexerState.Symbols.Add();
            var identifierSymbol = grammar.LexerStates.DefaultLexerState.Symbols.Insert(0,
                identifier + "Keyword", invalidVar, TerminalSymbolComparison.RegularExpression);
            //identifierSymbol.LexerStateToEnter.
            // Add the new ID as another alternate identifier that can be matched.
            alternateRule.Rules.Add(identifierSymbol);
            // Assign the color coding used by the identifier
            var languageElementName = identifier + "LanguageElement";
            identifierSymbol.LanguageElementName = languageElementName;

            if (ClassificationAppearanceMap == null)
                ClassificationAppearanceMap = new ClassificationAppearanceMap();

            ClassificationAppearanceMap.AddMapEntry(
                new ClassificationType(languageElementName, identifier + " Name", "The keyword for the " + identifier + " name"),
                new TextDocumentAppearance() { Foreground = new SolidColorBrush(foreground), FontBold = true });
        }

        public Grammar Grammar2()
        {
            var grammar = new Grammar();
            var defaultLexerState = grammar.LexerStates.DefaultLexerState;
            //grammar.LexerStates.
            // Include the built-in symbols which should be matched in the default lexer state.
            defaultLexerState.Symbols.Add(grammar.WhitespaceSymbol);
            defaultLexerState.Symbols.Add(grammar.NewLineSymbol);
            var equals = defaultLexerState.Symbols.Add("EqualsToken", "=");
            var percentage = defaultLexerState.Symbols.Add("PercentageToken", "%");
            var semicolon = defaultLexerState.Symbols.Add("SemicolonToken", ";");
            var varKeyword = defaultLexerState.Symbols.Add("VarKeyword", "var");
            //var identifier = defaultLexerState.Symbols.Add("IdentifierToken",
            //    "(?<TokenBegin>%)([_a-zA-Z0-9]*)(?<TokenEnd>%)", TerminalSymbolComparison.RegularExpression);
            //var type = grammar.NonTerminalSymbols.Add("Type");
            //type.Rule = identifier; // Implicit conversion
            //var identifier2 = defaultLexerState.Symbols.Add("IdentifierToken2",
            //    "((%LASTRUN%)|(%LASTRUN.M%)|(%LASTRUN.Y%)|(%LASTRUN.MMMM%))", TerminalSymbolComparison.RegularExpression);
            var identifier3 = defaultLexerState.Symbols.Add("IdentifierToken3",
                "((?i)(%LASTRUN%)|(%LASTRUN.SS%)|(%LASTRUN.YY%)|(%LASTRUN.MMMM%))(?-i)", TerminalSymbolComparison.RegularExpression);
            //var identifier4 = defaultLexerState.Symbols.Add("IdentifierToken4",
            //    "((?i)(%LASTRUN%))(?-i)", TerminalSymbolComparison.RegularExpression);
            //var type = grammar.NonTerminalSymbols.Add("Type");
            //var expression = grammar.NonTerminalSymbols.Add("Expression");
            //expression.Rule = identifier2;
            var variableDeclarationStatement =
                grammar.NonTerminalSymbols.Add("VariableDeclarationStatement");

            variableDeclarationStatement.Rule = //new EmptySyntaxRule();
            new ConcatenationSyntaxRule(
                percentage,
                identifier3,
                percentage
            );

            grammar.StartSymbol = variableDeclarationStatement;
            return grammar;
        }

        public Grammar CreateGrammar3()
        {
            var tsqlLanguage = TSqlLanguage.Instance;
            var customGrammar = tsqlLanguage.Grammar.Clone();
            var defaultLexerState = customGrammar.LexerStates.DefaultLexerState;

            NonTerminalSymbol identifierResolved;


            if (customGrammar.NonTerminalSymbols.TryGetSymbol("_identifierResolved", out identifierResolved) == false)
            {
                Debug.Fail("Cannot find the _identifierResolved symbol.");
                return null;
            }

            var alternateRule = identifierResolved.Rule as AlternationSyntaxRule;
            if (alternateRule == null)
            {
                Debug.Fail("The _identifierResolved symbol does not have an alternation at the root level rule, as expected.");
                return null;
            }

            var percentage = defaultLexerState.Symbols.Insert(0, "PercentageToken", "%");
            alternateRule.Rules.Add(percentage);

            var sid = "IdentifierToken3";
            var identifier3 = defaultLexerState.Symbols.Insert(0, sid,
                "((LASTRUN)|(LASTRUN.M)|(LASTRUN.YY)|(LASTRUN.MMMM))", TerminalSymbolComparison.RegularExpression);

            var variableDeclarationStatement =
                customGrammar.NonTerminalSymbols.Add("VariableDeclarationStatement");

            variableDeclarationStatement.Rule = new ConcatenationSyntaxRule(
                percentage,
                identifier3,
                percentage

            );

            // Assign the color coding used by the identifier
            var languageElementName = sid + "LanguageElement";
            identifier3.LanguageElementName = languageElementName;

            customGrammar.StartSymbol = variableDeclarationStatement;

            if (ClassificationAppearanceMap == null)
                ClassificationAppearanceMap = new ClassificationAppearanceMap();

            ClassificationAppearanceMap.AddMapEntry(
                new ClassificationType(languageElementName, sid + " Name", "The keyword for the " + sid + " name"),
                new TextDocumentAppearance() { Foreground = new SolidColorBrush(Colors.Red), FontBold = true });

            return customGrammar;

        }

        //public void ModLexerStates(Grammar grammar, string ValidVarRegEx)
        public void ModLexerStates(Grammar grammar, string ValidVarRegEx)
        {

            #region hide region

            NonTerminalSymbol identifierResolved;
            if (grammar.NonTerminalSymbols.TryGetSymbol("_identifierResolved", out identifierResolved) == false)
            {
                Debug.Fail("Cannot find the _identifierResolved symbol.");
                return;
            }

            var alternateRule = identifierResolved.Rule as AlternationSyntaxRule;
            if (alternateRule == null)
            {
                Debug.Fail("The _identifierResolved symbol does not have an alternation at the root level rule, as expected.");
                return;
            }

            #endregion

            var languageName = "Var";
            var defaultLexerState = grammar.LexerStates.DefaultLexerState;

            var varStart = defaultLexerState.Symbols.Add("VarStart", "/%");
            varStart.LanguageElementName = languageName;

            var varLexerState = grammar.LexerStates.Add("Var");
            varStart.LexerStateToEnter = varLexerState;

            //var varContent1 = varLexerState.Symbols.Add(
            //    "Var1", "[^%]+", TerminalSymbolComparison.RegularExpression);
            //varContent1.LanguageElementName = languageName;

            var varContent = varLexerState.Symbols.Add(
                "VarContent",
                //"(?i)((LASTRUN)|(LASTRUN.SS)|(LASTRUN.YY)|(LASTRUN.MMMM))(?-i)", TerminalSymbolComparison.RegularExpression);
                "(?i)(%LASTRUN%)(?-i)", TerminalSymbolComparison.RegularExpression);

            varContent.LanguageElementName = languageName;

            //var invalidVarMid = varLexerState.Symbols.Add(
            //    "VarMid", "%+", TerminalSymbolComparison.RegularExpression);
            //invalidVarMid.LookaheadPattern = "/";
            //invalidVarMid.LanguageElementName = languageName;

            var invalidVarEnd = varLexerState.Symbols.Add(
                "VarEnd", "%/", isExitSymbol: true);
            invalidVarEnd.LanguageElementName = languageName;

            if (ClassificationAppearanceMap == null)
                ClassificationAppearanceMap = new ClassificationAppearanceMap();

            ClassificationAppearanceMap.AddMapEntry(
                new ClassificationType(languageName, varStart + " Name", "The keyword for the " + varStart + " name"),
                new TextDocumentAppearance() { Foreground = new SolidColorBrush(Colors.Chartreuse), FontBold = true });


        }

        //old code and commented code below
        //private void CreateCustomGrammar(List<string> invalidVars)
        //{

        //    try
        //    {
        //        //if (!QueryText.GetVariables().Any())
        //        //    return;

        //        ClassificationAppearanceMap = null;

        //        var tsqlLanguage = TSqlLanguage.Instance;
        //        var customGrammar2 = tsqlLanguage.Grammar.Clone();
        //        DAVariableLanguage(customGrammar2);
        //        // var customGrammar = Grammar2();
        //        //var customGrammar = new Grammar();
        //        //ModLexerStates(customGrammar, CreateValidVarRegEx());
        //        //var customGrammar = ModLexerStates2();

        //        //var n1 = @"(\b)((LASTRUN)|(LASTRUN.M)|(LASTRUN.Y)|(LASTRUN.MMMM))(\b)";
        //        //var n1 = @"((LASTRUN)|(LASTRUN.M)|(LASTRUN.Y)|(LASTRUN.MMMM))";
        //        //var n1 = @"q(?!u|r)";
        //        //var n1 = @"[^(?i)(?<TokenBegin>%)(?<Variable>LASTRUN|LASTRUN.M)(?<TokenEnd>%)(?-i)]";
        //        //var n1 = @"(?i)(?<TokenBegin>%)(?<Variable>LASTRUN|LASTRUN.M)(?<TokenEnd>%)(?-i)";

        //        //AddIdentifierSymbol(customGrammar, "InvalidVar", n1, Colors.Red);

        //        //int j = 0;
        //        //invalidVars.ForEach(x =>
        //        //{
        //        //    AddIdentifierSymbol(customGrammar, "Error" + j++.ToString(), "%" + x + "%", Colors.Red);
        //        //});

        //        //should be list of valid vars split by pipes
        //        //ar n2 = @"[^%LASTRUN%|%LASTRUN.MM%|%LASTRUN.D%]";
        //        //var n3 = @")++)*/%.*";

        //        //mobile(?:/(?:(?!sale|search|lot|login)[^/])++)*/retail.*

        //        var customLanguage = new CustomLanguage(customGrammar2);


        //        //foreach (var serviceKey in tsqlLanguage.ServicesManager.GetRegisteredServiceKeys())
        //        //    customLanguage.ServicesManager.RegisterService(serviceKey, tsqlLanguage.ServicesManager.GetService(serviceKey));

        //        //foreach (var serviceType in tsqlLanguage.ServicesManager.GetRegisteredServiceTypes())
        //        //    customLanguage.ServicesManager.RegisterService(serviceType, tsqlLanguage.ServicesManager.GetService(serviceType));

        //        // re-building document here prevents drag drop error here - listbox_PreviewMouseLeftButtonDown
        //        //can potentially eventually store caret position as well, but re-creating language this way on every keystroke seems unsustainable
        //        //var holdText = TextDocument.CurrentSnapshot.GetText();
        //        //_textDocument = new TextDocument();
        //        //_textDocument.AutoCasingMode = TextDocumentAutoCasingMode.None;
        //        //_textDocument.TextChanged += TdOnTextChanged;
        //        //_textDocument.InitializeText(holdText);
        //        //RaisePropertyChanged(() => TextDocument);
        //        //

        //        TextDocument.Language = customLanguage;
        //    }
        //    catch (Exception e)
        //    {

        //    }

        //}


    }
}
